import base64
import json
import os
from io import StringIO
from typing import Dict, Type, Any

import numpy as np
import pandas as pd
from pykrige import OrdinaryKriging
from sklearn.linear_model import LinearRegression

from landsklim.lisdqs.cache import qgis_project_cache
from landsklim.lisdqs.landsklim_constants import DATASET_RESPONSE_VARIABLE
from landsklim.lisdqs.lisdqs_analysis import LISDQSAnalysis
from landsklim.lisdqs.lisdqs_configuration import LISDQSConfiguration
from landsklim.lisdqs.lisdqs_interpolation import LandsklimRectangle, LISDQSInterpolation, LISDQSInterpolationType
from landsklim.lisdqs.lisdqs_project import LISDQSProject
from landsklim.lisdqs.map_layer import MapLayer, VectorLayer, RasterLayer, MapLayerCollection
from landsklim.lisdqs.phase import IPhase
from landsklim.lisdqs.phase_composite import PhaseComposite
from landsklim.lisdqs.phase_multiple_regression import PhaseMultipleRegression
from landsklim.lisdqs.polygons_definition import PolygonsDefinition
from landsklim.lisdqs.regression_model import MultipleRegressionModel
from landsklim.lisdqs.regressor import Regressor
from landsklim.serialization.json_encoder import PUBLIC_ENUMS
from landsklim.serialization.landsklim_unpickler import PearsonRResultDummyClass


class LandsklimDecoder(json.JSONDecoder):
    """
    Deserialize Landsklim classes from JSON
    """

    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)
        self.lk_class_map = {
            'np.int': np.integer,
            'np.float': np.floating,
            "LandsklimRectangle": LandsklimRectangle,
            'PearsonRResultDummyClass': PearsonRResultDummyClass,
            'LinearRegression': LinearRegression,
            'OrdinaryKriging': OrdinaryKriging,
            'LISDQSProject': LISDQSProject,
            'LISDQSConfiguration': LISDQSConfiguration,
            'LISDQSAnalysis': LISDQSAnalysis,
            'MultipleRegressionModel': MultipleRegressionModel,
            'LISDQSInterpolation': LISDQSInterpolation,
            'PolygonsDefinition': PolygonsDefinition
        }
        subclasses = Regressor.__subclasses__() + IPhase.__subclasses__()
        for sc in subclasses:
            self.lk_class_map[sc.__name__] = sc

        maplayer_subclasses = MapLayer.__subclasses__()
        self.maplayer_class_map = {}
        for sc in maplayer_subclasses:
            self.maplayer_class_map[sc.__name__] = sc

    def rebuild_layer(self, dct: Dict, map_layer: MapLayer):
        map_layer._id = dct["id"]
        map_layer._lisdqs_code = dct["lisdqs_code"]
        map_layer.resolve()


    def __list_to_maplayercollection(self, lst: list):
        mlc = MapLayerCollection()
        for l in lst:
            mlc.append(l)
        return mlc

    def rebuild_maplayercollection(self, instance: LISDQSProject):
        instance._stations = self.__list_to_maplayercollection(instance._stations)
        instance._variables = self.__list_to_maplayercollection(instance._variables)
        instance._additional_layers = self.__list_to_maplayercollection(instance._additional_layers)

    def get_json_dependencies(self, analysis: LISDQSAnalysis):
        path = os.path.join(qgis_project_cache().homePath(), os.path.normpath(analysis.json_polygons).replace("\\", "/"))
        with open(path, "rb") as json_file:
            analysis._polygons = json.load(json_file, cls=LandsklimDecoder)
        delattr(analysis, "json_polygons")

        path = os.path.join(qgis_project_cache().homePath(), os.path.normpath(analysis.json_models).replace("\\", "/"))
        with open(path, "rb") as json_file:
            analysis._models = json.load(json_file, cls=LandsklimDecoder)
        delattr(analysis, "json_models")

    def rebuild_hierarchy(self, instance: object):
        """
        Recreate cyclic links in the object graph between
         LISDQSProject, LISDQSConfiguration, LISDQSAnalysis and LISDQSInterpolation
        """
        if isinstance(instance, LISDQSProject):
            for configuration in instance.get_configurations():
                configuration._lisdqs_project = instance
        if isinstance(instance, LISDQSConfiguration):
            for analysis in instance.get_analysis():
                analysis._configuration = instance
        if isinstance(instance, LISDQSAnalysis):
            for interpolation in instance.get_interpolations():
                interpolation._analysis = instance

    def postprocessing(self, instance:  Any):
        """
        Correct by hand what can't be deserialized by the JSONDecoder
        Especially :
         - Rebuild the hierarchy in the object graph
         - Rebuild MapLayerCollection objects from LISDQSProject
         - Convert dicts string keys to int keys when needed
         - IPhase recovers polygons definition from LISDQSAnalysis which they have dropped
         - PhaseComposite recovers phase_1 and phase_2 which they have dropped
        """
        self.rebuild_hierarchy(instance)

        # OrdinaryKriging.variogram_function is a lambda which cannot be serialized
        if isinstance(instance, OrdinaryKriging):
            instance.variogram_function = instance.variogram_dict[instance.variogram_model]

        if isinstance(instance, LISDQSProject):
            self.rebuild_maplayercollection(instance)

        # The nested dict keys are regressors,
        # which were to be converted into their index to be serialized : rebuild the dict
        if isinstance(instance, LISDQSAnalysis):
            self.get_json_dependencies(instance)

            rebuild_pc = {}
            for key, value in instance._pearson_correlation.items():
                rebuild_pc[int(key)] = {}
                for reg_index, pc in value.items():
                    reg = instance._regressors[int(reg_index)]
                    rebuild_pc[int(key)][reg] = pc
            instance._pearson_correlation = rebuild_pc

            new_datasets = {}
            for k, v in instance._datasets.items():
                new_datasets[int(k)] = v
            instance._datasets = new_datasets

            new_kriging_layers = {}
            for k, v in instance._kriging_layers.items():
                new_kriging_layers[int(k)] = v
            instance._kriging_layers = new_kriging_layers

            new_models = {}
            for k, v in instance._models.items():
                new_models[int(k)] = v
            instance._models = new_models

            # Repopulate IPhase polygons from LISDQSAnalysis
            for situation in instance.get_station_situations():
                for phase in instance.get_phases(situation):
                    phase._polygons = instance.get_polygons()

            # Refill PhaseComposite phases
            for situation in instance.get_station_situations():
                for phase in instance.get_phases(situation):
                    if isinstance(phase, PhaseComposite):
                        phase._phases = [instance.get_phases(situation)[0], instance.get_phases(situation)[1]]

        if isinstance(instance, IPhase):
            new_corr = {}
            for k, v in instance._variables_correlation.items():
                new_corr[k] = {}
                for vk, vv in v.items():
                    new_corr[k][int(vk)] = vv
            instance._variables_correlation = new_corr

        if isinstance(instance, LISDQSInterpolation):
            new_layers = {}
            for k, v in instance._layers.items():
                _, member = k.split('.')  # format of str is 'LISDQSInterpolationType.VALUE'
                nk = getattr(LISDQSInterpolationType, member)
                new_layers[nk] = {}
                for vk, vv in v.items():
                    new_layers[nk][int(vk)] = vv
            instance._layers = new_layers

        # Restore scale_means and scale_std in MultipleRegressionModel objects
        if isinstance(instance, PhaseMultipleRegression):
            snames = instance.scale_column_names
            for model, (_, smeans), (_, sstd), dmins, dmaxs in zip(instance._model, instance.scale_means.iterrows(), instance.scale_std.iterrows(), instance.dataset_mins, instance.dataset_maxs):
                setattr(model, "_scale_means", smeans)
                setattr(model, "_scale_std", sstd)
                setattr(model, "_scale_column_names", snames)
                setattr(model, "_dataset_mins", dmins)
                setattr(model, "_dataset_maxs", dmaxs)

                # MultipleRegressionModel._pearson_correlations are serialized as array (and not dict)
                new_pc = {}
                for i, c in enumerate(snames):
                    if c != DATASET_RESPONSE_VARIABLE:
                        new_pc[c] = [model.pearson_correlations[2*i], model.pearson_correlations[2*i+1]]
                setattr(model,"_pearson_correlations", new_pc)
                delattr(model, "pearson_correlations")


            delattr(instance, "scale_means")
            delattr(instance, "scale_std")
            delattr(instance, "scale_column_names")
            delattr(instance, "dataset_mins")
            delattr(instance, "dataset_maxs")

        # MultipleRegressionModel._predictors_are_integers has been removed on serialisation
        # MultipleRegressionModel._standardize has been removed on serialisation
        if isinstance(instance, MultipleRegressionModel):
            setattr(instance, "_predictors_are_integers", False)
            setattr(instance, "_standardize", MultipleRegressionModel.STANDARDIZE_DEFAULT_VALUE)
            setattr(instance, "_MultipleRegressionModel__residuals", None)
            setattr(instance, "debug", -1)


    def object_hook(self, dct: Dict) -> object:
        """
        Each parsed JSON dictionary is sent to this method.
        """

        if '__class__' in dct:
            cls_item = dct['__class__']
            if cls_item == 'tuple':
                return tuple(dct['array'])
            elif cls_item == "np.array":
                return np.array(dct['array'], dtype=dct['type'])
            elif cls_item == 'pd.DataFrame':
                return pd.read_csv(StringIO(dct['data']), index_col=0)
                # return pd.read_json(dct['data'], orient='split')
            elif cls_item == 'pd.Series':
                dct.pop('__class__')
                dtype = dct.pop('type')
                return pd.Series(dct, dtype=dtype)
            elif cls_item in self.maplayer_class_map:
                cls: Type = self.maplayer_class_map[cls_item]
                instance: MapLayer = cls.__new__(cls)
                self.rebuild_layer(dct, instance)
                return instance
            else:  # Every other types with a serialized __class__ attribute is a "Landsklim object"
                cls: Type = self.lk_class_map[cls_item]
                instance: object =  self.__new_instance(cls, dct)
                self.postprocessing(instance)
                return instance

        if '__enum__' in dct:
            enum_value: str = dct['__enum__']
            enum_name, member = enum_value.split('.')  # format of __enum__ is 'ENUM_CLASS.VALUE'
            return getattr(PUBLIC_ENUMS[enum_name], member)
        return dct

    def __new_instance(self, cls: Type, state: Dict[str, Any]) -> Any:
        """
        Create an instance of any class by passing its type and its state

        :param cls: Type of the object to instantiate
        :type cls: Type

        :param state: The attributes of the object to be instantiated
        :type: state: Dict[str, Any]

        :returns: A new object
        :rtype: Any

        """
        state.pop('__class__')  # Remove the __class__ attribute who was there to identify the type of instantiate
        obj = cls.__new__(cls)
        for attr, value in state.items():
            setattr(obj, attr, value)
        return obj
